import{MenuController}from"../menu.mjs";class DocumentationController extends MenuController{static helpLink="https://github.com/painebenjamin/app.enfugue.ai/wiki";static menuName="Documentation";static menuIcon="fa-solid fa-book";async onClick(){window.open(this.constructor.helpLink)}}export{DocumentationController as MenuController};
