let Configuration={debug:!0,controller:{module:"index"},view:{applicationContainer:"enfugue-application",language:"en"},url:{root:"http://dev.enfugue.local/",api:"http://dev.enfugue.local/api/",baseTitle:"Enfugue",titleSeparator:" | "},history:{size:100},model:{cache:0,cookie:{name:"enfugue_token"},status:{interval:1e4},autosave:{interval:3e4},queue:{interval:5e3},downloads:{interval:5e3},invocation:{width:512,height:512,chunkingSize:64,chunkingBlur:64,guidanceScale:7.5,inferenceSteps:50,interval:1e3,upscaleDiffusionSteps:80,upscaleDiffusionGuidanceScale:10,upscaleDiffusionStrength:.15,upscaleDiffusionChunkingSize:64,upscaleDiffusionChunkingBlur:64,errors:{consecutive:2}}}};export{Configuration};
