import{ElementBuilder}from"../base/builder.mjs";const E=new ElementBuilder,defaultDuration=5e3,hideDuration=1e3;class SimpleNotification{static notify(e,t){let i=E.span().class("notification").content(e),o=E.div().class("notification-container").content(i);document.body.appendChild(o.render()),setTimeout((()=>{o.css("opacity",0),setTimeout((()=>{document.body.removeChild(o.element)}),1e3)}),void 0===t?5e3:t)}}export{SimpleNotification};
