import{MenuController}from"../menu.mjs";class NewModelController extends MenuController{static menuName="New Model";static menuIcon="fa-solid fa-plus";async onClick(){this.application.modelManager.showNewModel()}}export{NewModelController as MenuController};
