import{Application}from"./application/index.mjs";import{Loader}from"./base/loader.mjs";import{getQueryParameters,isEmpty}from"./base/helpers.mjs";(async()=>{let i,o,e=getQueryParameters(),t=window.enfugue||{},n="index";if(!isEmpty(e.config))try{o=await import(`./config/${e.config}.mjs`),i=e.config}catch(i){console.error("Couldn't get configuration",e.config,", using default.")}void 0===o&&(i=n,o=await import(`./config/${n}.mjs`));let r=o.Configuration;r.name=i,isEmpty(t.url)||(r.url={...r.url||{},...t.url}),isEmpty(t.keys)||(r.keys=t.keys);const a=new Application(r);Loader.done((function(){a.initialize(),e.debug&&console.log("Application initialized",a)}))})();
