import{Model,ModelObject}from"./index.mjs";class DiffusionModel extends ModelObject{static alwaysInclude=!0;static apiRoot="models";static apiScope=["name"];static apiInclude=["lora","inversion"];getTensorRTStatus(){return this.queryModel("get",`${this.url}/tensorrt`)}}class DiffusionModelInversion extends ModelObject{}class DiffusionModelLora extends ModelObject{}class DiffusionInvocation extends ModelObject{static apiRoot="invocation-history";static apiScope=["id"]}class User extends ModelObject{static alwaysInclude=!0;static apiRoot="users";static apiScope=["username"];static apiInclude=["permission_groups","permission_groups.group"]}class UserPermission extends ModelObject{}class UserPermissionGroup extends ModelObject{}class PermissionGroup extends ModelObject{}class EnfugueModel extends Model{static modelObjects=[DiffusionModel,DiffusionModelInversion,DiffusionModelLora,DiffusionInvocation,User,UserPermission,UserPermissionGroup,PermissionGroup]}export{EnfugueModel as Model,ModelObject};
