import{InputView}from"./base.mjs";class StringInputView extends InputView{static maxLength=128;static pattern=null;static minLength=null;async build(){let t=await super.build();return t.attr("maxlength",this.constructor.maxLength),t.attr("minlength",this.constructor.minLength),t.attr("pattern",this.constructor.pattern),t}}class TextInputView extends InputView{static rows=5;static tagName="textarea";static maxLength=1024;async build(){let t=await super.build();return t.attr("rows",this.constructor.rows).attr("maxlength",this.constructor.maxLength),t}}class PasswordInputView extends StringInputView{static inputType="password";static autoComplete="new-password"}export{StringInputView,TextInputView,PasswordInputView};
