import{isEmpty}from"../base/helpers.mjs";class Controller{static confirmViewWidth=500;static confirmViewHeight=200;constructor(i){this.application=i}async initialize(){}get model(){return this.application.model}get images(){return this.application.images}get config(){return this.application.config}get engine(){return this.application.engine}get downloads(){return this.application.downloads}notify(i,t,n){return this.application.notifications.push(i,t,n)}spawnWindow(){return this.application.windows.spawnWindow.apply(this.application.windows,Array.from(arguments))}download(i,t,n){return this.downloads.download(i,t,n)}confirm(i){return this.application.confirm(i)}subscribe(i,t){this.application.subscribe(i,t)}publish(i,t=null){this.application.publish(i,t)}getState(){return{}}getDefaultState(){return{}}setState(i){}}export{Controller};
