import{MenuController}from"../menu.mjs";class NewFileController extends MenuController{static menuName="New";static menuIcon="fa-solid fa-file";async onClick(){await this.application.resetState()}}export{NewFileController as MenuController};
