import{MenuController}from"../menu.mjs";import{createEvent,isEmpty}from"../../base/helpers.mjs";import{FormView}from"../../view/forms/base.mjs";import{StringInputView}from"../../view/forms/input.mjs";class FileNameFormView extends FormView{static fieldSets={"File Name":{filename:{class:StringInputView,config:{required:!0,value:"Enfugue Project"}}}}}class SaveFileController extends MenuController{static menuName="Save As";static menuIcon="fa-solid fa-file-arrow-down";async onClick(){this.application.saveAs("Save Project",JSON.stringify(this.application.getState()),"application/json",".json")}}export{SaveFileController as MenuController};
