import{Language as English}from"./index.mjs";English.strings.common={all:"All",submit:"Submit",reset:"Reset",error:"Please correct the above errors.",selectOne:"Select One"};export{English as Language};
