import{InputView}from"./base.mjs";class CheckboxInputView extends InputView{static inputType="checkbox";setValue(e,t){return"boolean"!=typeof e&&(e="string"==typeof e?-1!==["true","t","yes","y"].indexOf(e.toLowerCase()):!!e),super.setValue(e,t)}getValue(){return void 0!==this.node&&void 0!==this.node.element?this.node.element.checked:!!this.value}toggleValue(){this.disabled||this.setValue(!this.getValue())}labelClicked(){this.toggleValue()}}export{CheckboxInputView};
