import{InputView}from"./base.mjs";class HiddenInputView extends InputView{static inputType="hidden"}class ButtonInputView extends InputView{static inputType="button";getValue(){return this.value}setValue(t,e){this.value=t,e&&this.changed()}async build(){let t=this.value;this.value=this.constructor.defaultValue;let e=await super.build();return this.value=t,e.off("click").on("click",(t=>this.changed(t))),e}}export{HiddenInputView,ButtonInputView};
