import{isEmpty}from"../base/helpers.mjs";import{Controller}from"./base.mjs";class MenuController extends Controller{static menuName;static menuIcon;static menuShortcut;static isDisabled(){return isEmpty(this.menuName)}async onClick(){this.notify("warn","Unimplemented Menu Controller","Override the onClick() function to give this menu controller an implementation.")}}export{MenuController};
